// DRAW.CPP

#include <graphics.h>
#include <mem.h>
#include <stdlib.h>
#include "music.h"

// ICONS
extern char gclef[], qrest[], flat[], sharp[], natural[];

enum
{
STAFFYSIZE=90,
STAFFYOFF=95,

MINSPACING=40,
FIRSTCOUNTERX=35,
LASTCOUNTERX=635,

BARXSIZE=10,
THICKBARXSIZE=13,
BUFFERZONE=30,
TIMEXSIZE=20,

STEMSIZE=30,
MINSTEMSIZE=25,

CLXSIZE=8,
};

enum
{
nbfNUMFLAGS=3,

nbfRUNSTATE=0x0C,
nbfNORUN=0,
nbfFIRSTRUN=4,
nbfMIDDLERUN=8,
nbfLASTRUN=12,

nbfTRIPLES=0x30,

nbfTIE=0x40,

nbfEOM=0x80,
};

struct notebuf_t
{
	int notey, sticky, dir;
	int flags;

	int numflags() const { return (int)(flags&0x03); }
	int runstate() const { return (int)(flags&0x0C); }
	int triples() const { return (int)((flags>>4)&0x03); }
	int tied() const { return (int)(flags&nbfTIE); }
	int eom() const { return (int)(flags&nbfEOM); }
};

//-------------------------------------------------------
int staffy, xcounter;
int notex, spacing, notein;

int notey, sticky, stickx, dir;

int h_mc, h_measure;
int* h_notec;

int tx1, ty1, tx2, ty2, txm, tdy;
int fermata;

signed char keysig[7];
char caccs[64];
char keyorder[]={5,2,6,3,0,4,1};
int keyoff[]={65,50,35,55,40,60,45};

notebuf_t nb[(LASTCOUNTERX-FIRSTCOUNTERX)/MINSPACING+1];
int nbc;

int flagdown[5][2]=	{4,23,	7,10,	1,0,	1,3,	7,10};
int flagup[5][2]=	{4,0,	7,13,	1,23,	1,20,	7,13};

int barflag[5][2];
char numflags[ALLNOTETYPES]={3,2,1,0,0,0,2,1,0,0,0,1,0,1,0};

signed char numtable[]={
-23,-22,-21,-20,-19,-18,-17,0,
-16,-15,-14,-13,-12,-11,-10,0,
-9,-8,-7,-6,-5,-4,-3,0,
-2,-1,0,1,2,3,4,0,
5,6,7,8,9,10,11,0,
12,13,14,15,16,17,18,0,
19,20,21,22,23,24,25,0,
26,27,28,29,30,31,32,0,
33,34,35,36,37,38,39,0
};

//-------------------------------------------------------
/**/ inline
void drawbar(int x)
{
 line(x,staffy-50,x,staffy-10);
}

/**/ inline
void drawthickbar(int x)
{
 bar(x,staffy-50,x+4,staffy-10);
}

void drawwholenote()
{
 ellipse(notex+6,notey,0,360,6,5);
 ellipse(notex+6,notey,0,360,6,4);
}

void drawhalfnote()
{
 drawwholenote();
 line(stickx,notey,stickx,sticky);
}

void drawquarternote()
{
 fillellipse(notex+6,notey,6,5);
 line(stickx,notey,stickx,sticky);
}

void drawbridge(int i)
{
register int k, l;

 k=i*8*dir;
 l=5*dir;

 barflag[4][0]=barflag[1][0]=barflag[0][0]=stickx;
 barflag[2][0]=barflag[3][0]=stickx+spacing;
 barflag[4][1]=barflag[0][1]=sticky-k;
 barflag[1][1]=barflag[0][1]-l;
 barflag[3][1]=nb[nbc+1].sticky-k;
 barflag[2][1]=barflag[3][1]-l;

 fillpoly(5,barflag[0]);
}

void drawcantilever(register int i, register int lrdir)
{
 barflag[4][0]=barflag[1][0]=barflag[0][0]=stickx;
 barflag[2][0]=barflag[3][0]=stickx+(lrdir?-CLXSIZE:+CLXSIZE);
 barflag[4][1]=barflag[3][1]=barflag[0][1]=sticky-i*8*dir;
 barflag[2][1]=barflag[1][1]=barflag[0][1]-5*dir;

 fillpoly(5,barflag[0]);
}

void draweighthnote()
{
register int i, j;
int kt, kn, kp;

 kt=nb[nbc].numflags();

 drawquarternote();
 switch((nb[nbc].flags&nbfRUNSTATE)>>2)
	{
	case nbfNORUN>>2:
	 if(sticky<notey) for(i=0;i<kt;i++)
		{
		 setviewport(stickx,j+3,stickx+10,(j=sticky+i*6)+26,0);
		 drawpoly(5,flagdown[0]);
		 setviewport(0,0,639,479,0);
		}
	 else for(i=0;i<kt;i++)
		{
		 setviewport(stickx,j-26,stickx+10,(j=sticky-i*6)+3,0);
		 drawpoly(5,flagup[0]);
		 setviewport(0,0,639,479,0);
		}
	 break;

	case nbfFIRSTRUN>>2:
	 kn=nb[nbc+1].numflags();
	 j=kt<kn?kt:kn;
	 for(i=0;i<j;i++) drawbridge(i);
	 for(;i<kt;i++) drawcantilever(i,0);
	 break;

	case nbfMIDDLERUN>>2:
	 kn=nb[nbc+1].numflags();
	 j=kt<kn?kt:kn;
	 for(i=0;i<j;i++) drawbridge(i);
	 if(i<kt)
		{
		 kp=nb[nbc-1].numflags();
		 if(kp>kn) { j=kp; kn=0; }
		 else { j=kn; kn=1; }
		 for(i=j;i<kt;i++)
			drawcantilever(i,kn);
		}
	 break;

	case nbfLASTRUN>>2:
	 kp=nb[nbc-1].numflags();
	 if(kp<kt) for(i=kp;i<kt;i++) drawcantilever(i,1);
	 break;
	}
}

void draweighthrest(register int x, register int y)
{
 line(x+3,y+20,x+7,y);
 line(x,y+8,x+6,y+5);
 fillellipse(x,y+6,2,2);
}

void drawthree(int x, int y)
{
 settextstyle(TRIPLEX_FONT,HORIZ_DIR,1);
 settextjustify(CENTER_TEXT,CENTER_TEXT);
 outtextxy(x,y,"3");
}

//-------------------------------------------------------
int interpolate(register int x1, register int y1, int x2, int y2, int xm)
{
 return ((xm-x1)*(y2-y1))/(x2-x1)+y1;
}

int isrunnable(note_t far* n)
{
register int c;

 if(n->_name==REST || (c=n->_dur)>=nnCONTROL) return 0;
 return (int)numflags[c&nnDURATION];
}

void runipa(int f, register int i, int l, int dir)
{
int j;
register int k;
 nb[i].sticky=k=interpolate(f,nb[f].sticky,l,nb[l].sticky,i);

 if(dir==-1) 
	{
	 if((j=nb[i].notey-k-MINSTEMSIZE)<0)
		for(k=f;k<=l;k++) nb[k].sticky+=j;
	}
 else
	{
	 if((j=k-nb[i].notey-MINSTEMSIZE)<0)
		for(k=f;k<=l;k++) nb[k].sticky-=j;
	}
}

int inthypot(int x, int y)
{
register int r, rsq=x*x+y*y;
 for(r=x>y?x:y;;r++) if(r*r>=rsq) break;
 return r;
}

// this is a little algorithm to calculate the arctangent of a rational
// number using integer math.  I used the third taylor polynomial of the
// arctangent function,  T3(x) = x-x^3/3, and converted the result to
// degrees.  I apporximated pi by 22/7
// I simplified the expression for added speed
// it is fairly accurate for 40 degrees
int intarctan(register int dx, register int dy)
{
 long dx2=dx*dx;
 long dy3=dy*dy*dy;
 return -(1260L*(3*dy*dx2-dy3))/(66*dx2*dx);
}

void onenotetie(int sx1, int sy1, int sx2, int sy2)
{
int sxm=(sx1+sx2)/2, sym=(sy1+sy2)/2;
int dx=sx2-sx1, dy=sy2-sy1;

 if(dy>dx) return;

register int d=inthypot(abs(dx),abs(dy));
int a=(d*100)/83;
register int r=(d*25)/19;
int start=intarctan(dx,dy)+68, end=start+45;

 ellipse(sx1=sxm-(dy*a)/d,sy1=sym+(dx*a)/d,start,end,r,r);
 ellipse(sx1,sy1,start,end,r,r-1);
}

void newline(note_t far* n)
{
register int i;
int j;

int x, wx;

int cur, lm;
register int cbuf;

int eom;
int mc=h_mc, measure=h_measure;
int* notec=h_notec;

int run, lastdir;
int trip;

unsigned char dur, name;

 if(staffy>390)
	{
	 cleardevice();
	 settextstyle(DEFAULT_FONT,HORIZ_DIR,1);
	 settextjustify(LEFT_TEXT,TOP_TEXT);
	 outtextxy(10,470,program);
	 if(staffy==5000)
		{
		 settextstyle(TRIPLEX_FONT,HORIZ_DIR,4);
		 settextjustify(CENTER_TEXT,CENTER_TEXT);
		 outtextxy(320,30,csong->lines[0]);
		 settextstyle(TRIPLEX_FONT,HORIZ_DIR,2);
		 for(i=1;i<4;i++) if(csong->lines[i]!=NULL)
			outtextxy(320,52+i*18,csong->lines[i]);
		 staffy=STAFFYOFF+STAFFYSIZE;
		}
	 else staffy=STAFFYOFF;
	}
 else staffy+=STAFFYSIZE;

 for(i=staffy-50;i<staffy;i+=10) line(5,i,635,i);
 drawbar(5);
 putimage(7,staffy-62,gclef,OR_PUT);

 x=FIRSTCOUNTERX;

 for(i=0;i<7;i++)
	if(keysig[keyorder[i]]==1)
		putimage(x,staffy-keyoff[i],sharp,OR_PUT), x+=10;
 for(i=6;i>=0;i--)
	if(keysig[keyorder[i]]==-1)
		putimage(x,staffy-keyoff[i]-1,flat,OR_PUT), x+=10;
 wx=(xcounter=x)-FIRSTCOUNTERX;

 lm=cbuf=run=trip=0;
 for(cur=0;x<LASTCOUNTERX;cur++)
	{
	 dur=n[cur]._dur;
	 name=n[cur]._name;
	 if(dur>=nnCONTROL) switch(dur)
		{
		case ccFUNCTION:
		 switch(name)
			{
			case cfENDOFSONG:
			 lm=cbuf;
			 x+=THICKBARXSIZE, wx+=THICKBARXSIZE;
			 nb[cbuf-1].flags&=(~nbfEOM);
			 x=2000;

			 if(trip) nb[cbuf-trip].flags|=trip<<4, trip=0;
			 break;

			case cfENDOFMEASURE:
			 lm=cbuf;
			 x+=BARXSIZE, wx+=BARXSIZE;
			 nb[cbuf-1].flags|=nbfEOM;
			 mc=0;

			case cfRUNBREAK:
			 if(trip) nb[cbuf-trip].flags|=trip<<4, trip=0;
			 break;
			}
		 continue;

		case ccSETTIME:
		 measure=32*(name>>4);
		 switch(name&0x0F)
			{
			case 2: notec=fullnotedurs[0]; break;
			case 4: notec=fullnotedurs[1]; break;
			case 8: notec=fullnotedurs[2]; break;
			}
		 x+=TIMEXSIZE, wx+=TIMEXSIZE;
		 continue;

		default:
		 continue;
		}

// ADJUST X COUNTER, ETC.
	 x+=MINSPACING;

// CALCULATE NUBMER OF FLAGS IN FLAGGED NOTE OR REST
	 nb[cbuf].flags=numflags[dur&nnDURATION];

// CHECK TO SEE IF WE'VE REACHED THE END OF THE MEASURE
	 eom=(mc+=notec[dur&nnDURATION])>=measure;

// HANDLE RESTS
	 if(name==REST)
		{
		 nb[cbuf].dir=0;
		 nb[cbuf].notey=staffy-20;
		 nb[cbuf].sticky=staffy-40;
		 goto skipnotestuff;
		}

// CHECK TO EXTEND OR BEGIN RUNNING NOTE
	 if(run==0)
		{
		 if(!eom && isrunnable(n+cur) && isrunnable(n+cur+1))
			nb[cbuf].flags|=nbfFIRSTRUN, run=1;
		}
	 else
		{
		 if(!eom && isrunnable(n+cur+1) && run<3)
			nb[cbuf].flags|=nbfMIDDLERUN, run++;
		 else
			nb[cbuf].flags|=nbfLASTRUN, run+=11;
		}

// CALCULATE NOTEY, STICKY, AND STICKX FIELDS
	 nb[cbuf].notey=i=staffy-(j=numtable[name&nnNAMEOCT]*5);
	 if(run>1 || trip>0)
		nb[cbuf].dir=lastdir;
	 else
		nb[cbuf].dir=lastdir=(j<30?-1:1);
	 nb[cbuf].sticky=i+STEMSIZE*lastdir;

// INTERPOLATE RUNS (SMOOTH 'EM OUT)
	 if(run>10)
		{
		 j=0;
		 for(i=cbuf+11-run;i<=cbuf;i++) j+=nb[i].notey;
		 j=j/(run-10)-staffy;
		 if((lastdir==-1 && j<-28) || (lastdir==1 && j>-32))
			{
			 lastdir=-lastdir;
			 for(i=cbuf+11-run;i<=cbuf;i++)
				{
				 nb[i].dir=lastdir;
				 nb[i].sticky=lastdir*STEMSIZE+nb[i].notey;
				}
			}

		 if(run==13) runipa(cbuf-2,cbuf-1,cbuf,lastdir);
		 else if(run==14)
			{
			 runipa(cbuf-3,cbuf-2,cbuf,lastdir);
			 runipa(cbuf-3,cbuf-1,cbuf,lastdir);
			}

		 run=0;
		}

// CHECK FOR TRIPLES
	 if((dur&nnDURATION)>=twelfth)
		{ if(++trip==3) nb[cbuf-2].flags|=(3<<4), trip=0; }
	 else if(trip)
		nb[cbuf-trip].flags|=trip<<4, trip=0;

// CHECK FOR TIES
	 if(dur&nnTIE) nb[cbuf].flags|=nbfTIE;

skipnotestuff:
// IF WE'VE REACHED END OF MEASURE, PERFORM END OF MEASURE STUFF
	 if(eom)
		{
		 if(x<LASTCOUNTERX)	 // I'll fit it
			{
			 x+=BARXSIZE, wx+=BARXSIZE;
			 lm=cbuf+1;
			 mc=0;
			 nb[cbuf].flags|=nbfEOM;

			 if(trip) nb[cbuf-trip+1].flags|=trip<<4, trip=0;
			}
		}

// INCREMENT NOTE BUFFER COUNTER
	 cbuf++;
	}

 if(lm==0)	// measure can't fit on one line ...
	{
	 if((i=nb[cbuf-1].flags&nbfRUNSTATE)==nbfMIDDLERUN)
		nb[cbuf-1].flags+=4;
	 else if(i==nbfFIRSTRUN)
		nb[cbuf-1].flags-=4;
	 h_mc=mc;
	 spacing=MINSPACING;
	}
 else
	{
	 spacing=(LASTCOUNTERX-FIRSTCOUNTERX-wx)/lm;
	 if(spacing>100) spacing=100;
	}
 notein=(spacing-13)/2;
 nbc=0;
}

void drawnote(note_t far* n)
{
register int i;
int j, k;
register int c;
unsigned char dur, name;
char tc[2];

 dur=n->_dur;
 name=n->_name;
 if(dur>=nnCONTROL) switch(dur)
	{
	case ccFUNCTION:
	 switch(name)
		{
		case cfENDOFSONG:
		 drawbar(LASTCOUNTERX-7);
		 drawthickbar(LASTCOUNTERX-4);
		 return;

		case cfFERMATA:
		 fermata=1;
		 return;
		}
	 return;

	case ccSETSHARPFLAT:
	 switch(name&nnACCIDENTAL)
		{
		case nnFLAT: i=-1; break;
		case nnSHARP: i=1; break;
		case nnNATURAL: i=0; break;
		}
	 keysig[name&nnNAME]=i;
	 return;

	case ccSETTIME:
	 if(xcounter>LASTCOUNTERX-TIMEXSIZE) newline(n);

	 h_measure=32*(i=name>>4);
	 switch(j=name&0x0F)
		{
		case 2: h_notec=fullnotedurs[0]; break;
		case 4: h_notec=fullnotedurs[1]; break;
		case 8: h_notec=fullnotedurs[2]; break;			 
		}

	 settextstyle(TRIPLEX_FONT,HORIZ_DIR,4);
	 settextjustify(LEFT_TEXT,BOTTOM_TEXT);
	 tc[1]='\0';
	 tc[0]=i+'0';
	 outtextxy(xcounter+3,staffy-30,tc);
	 tc[0]=j+'0';
	 outtextxy(xcounter+3,staffy-10,tc);
	 xcounter+=TIMEXSIZE;
	 return;

	default:
	 return; 
	}

 if(xcounter>LASTCOUNTERX-spacing) newline(n);
 notex=xcounter+notein;
 c=nbc;

 if(name==REST)
	{
	 if(isdotted(k=dur&nnDURATION)) k-=subtractdot;
	 switch(k)
		{
		case whole:
		 bar(notex,staffy-40,notex+12,staffy-34);
		 break;

		case half:
		 bar(notex,staffy-36,notex+12,staffy-30);
		 break;

		case quarter:
		case sixth:
		case firstsixth:
		 putimage(notex+3,staffy-43,qrest,OR_PUT);
		 break;

		case eighth:
		case twelfth:
		case firsttwelfth:
		 draweighthrest(notex+3,staffy-40);
		 break;

		case sixteenth:
		 draweighthrest(notex+2,staffy-30);
		 draweighthrest(notex+4,staffy-40);
		 break;

		case thirtysecond:
		 draweighthrest(notex+1,staffy-30);
		 draweighthrest(notex+3,staffy-40);
		 draweighthrest(notex+5,staffy-50);
		 break;
		}

	 goto skipnotestuff;
	}

 notey=nb[c].notey;
 sticky=nb[c].sticky;
 dir=nb[c].dir;
 stickx=notex+(dir==-1?12:0);

 if(notey<=staffy-60) for(i=staffy-60;i>=notey;i-=10) line(notex-4,i,notex+16,i);
 if(notey>=staffy) for(i=staffy;i<=notey;i+=10) line(notex-4,i,notex+16,i);

 if(fermata)
	{
	 fillellipse(j=notex+6,i=staffy-60,2,2);
	 ellipse(j,i,0,180,10,9);
	 ellipse(j,i,0,180,10,10);
	 ellipse(j,i,0,180,10,11);
	 fermata=0;
	}

 switch(name&nnACCIDENTAL)
	{
	case nnFLAT:
	 if(caccs[name&nnNAMEOCT]!=1)
		putimage(notex-11,notey-16,flat,OR_PUT);
	 caccs[name&nnNAMEOCT]=1;
	 break;

	case nnSHARP:
	 if(caccs[name&nnNAMEOCT]!=2)
		putimage(notex-11,notey-15,sharp,OR_PUT);
	 caccs[name&nnNAMEOCT]=2;
	 break;

	case nnNATURAL:
	 if(caccs[name&nnNAMEOCT]!=3)
		 putimage(notex-10,notey-15,natural,OR_PUT);
	 caccs[name&nnNAMEOCT]=3;
	 break;
	}

 switch(dur&nnCONTROL)
	{
	case nnSTACCATO:
	 i=notey-dir*((notey-staffy)%10?10:15);
	 fillellipse(notex+6,i,2,2);
	 break;

	case nnLEGATO:
	 i=notey-dir*((notey-staffy)%10?10:15);
	 bar(notex,i-1,notex+12,i+1);
	 break;
	}

 if(isdotted(k=dur&nnDURATION))
	{		 
	 i=(notey-staffy)%10?notey:notey+5*dir;
	 fillellipse(notex+17,i,2,2);
	 k-=subtractdot;
	}
 switch(k)
	{
	case whole:
	 drawwholenote();
	 break;

	case half:
	 drawhalfnote();
	 break;

	case quarter:
	case sixth:
	case firstsixth:
	 drawquarternote();
	 break;

	case eighth:
	case sixteenth:
	case thirtysecond:
	case twelfth:
	case firsttwelfth:
	 draweighthnote();
	 break;	 
	}

 if(dur&nnTIE && ((!nb[c-1].tied() && !nb[c+1].tied()) 
	|| nb[c+1].notey==notey))
	{
	 onenotetie(notex+14,notey-4,notex+spacing+(nb[c].eom()?8:-2),
		nb[c+1].notey-4);
	}

skipnotestuff:
 if((j=nb[c].triples())>0)
	{
	 if(j==1) drawthree(notex+6,sticky+10*dir);
	 else
		{
		 for(i=0;i<j;i++) if(nb[c+i].runstate()==nbfNORUN) break;
		 k=(i!=j);

		 i=12*dir;
		
		 tx1=stickx-5;
		 ty1=sticky+i;
		 tx2=stickx+spacing*(--j)+5;
		 ty2=nb[c+j].sticky+i;
		 txm=(tx1+tx2)/2;
		 tdy=-10*dir;

// I might change this in my second version
/**/		 if(k)
/**/			{
/**/		 	 line(tx1,ty1,txm-8,interpolate(tx1,ty1,tx2,ty2,txm-8));
/**/			 line(txm+8,interpolate(tx1,ty1,tx2,ty2,txm+8),tx2,ty2);
/**/			 line(tx1,ty1,tx1,ty1+tdy);
/**/			 line(tx2,ty2,tx2,ty2+tdy);
/**/			}
/**/		 drawthree(txm,(ty1+ty2)/2);
		}
	}

 xcounter+=spacing;

 if(nb[c].flags&nbfEOM)
	{
	 i=xcounter+BARXSIZE/2;
	 if(LASTCOUNTERX-i<=BUFFERZONE) i=LASTCOUNTERX, xcounter=640;
	 else xcounter+=BARXSIZE;
	 drawbar(i);
	 memset(caccs,0,64);
	}

 nbc++; 
}

void initsong()
{
 xcounter=5000, staffy=5000;

 h_measure=128;
 h_mc=0;
 h_notec=fullnotedurs[1];

 settempo(120);

 ((int*)keysig)[0]=0;
 ((int*)keysig)[1]=0;
 ((int*)keysig)[2]=0;
 keysig[7]=0;

 memset(caccs,0,64);

 fermata=0;
}